#include "bp-cpu-integer.h"

// ********************************************************************
// Perform the Integer benchmark
// ********************************************************************

float run_integer(int num_threads, int run_time)
{
		pthread_t pt_monitor, pt_worker[num_threads];
		gchar tmp_str[200];
		if (num_threads == 1)
			sprintf(tmp_str, "Integer Benchmark (Single-threaded)");
		else
			sprintf(tmp_str, "Integer Benchmark (%d threads)", num_threads);
		gdk_threads_enter ();
		set_test_name(tmp_str);
		gdk_threads_leave ();

		pthread_mutex_lock (&mutex_reps);      
		reps = 0;
		pthread_mutex_unlock (&mutex_reps);      
		done = FALSE;
		
		if (pthread_create(&pt_monitor, NULL, sub_monitor, NULL) != 0)
			printf("ERROR: Could not create monitor thread\n");

    	start_timer_2();
    	
		int thread_cnt=0;
		for (thread_cnt=0; thread_cnt<num_threads; thread_cnt++)
		{
			if (pthread_create(&pt_worker[thread_cnt], NULL, thd_sub_all_integers, NULL) != 0)
				printf("ERROR: Could not create worker thread\n");		
		}
		
		long elapsed = get_time_2();
		float secs = (float)elapsed/10000000;

		sleep(run_time);
		done = TRUE;

		for (thread_cnt=0; thread_cnt<num_threads; thread_cnt++)
		{			
			pthread_join(pt_worker[thread_cnt], NULL);
		}
			
		elapsed = get_time_2();
		secs = (float)elapsed/10000000;

		pthread_join(pt_monitor, NULL);
		
		float result = (float)reps/secs;
		
		return result;
		
}

// ********************************************************************
// Thread calls each test once
// ********************************************************************
void *thd_sub_all_integers(void *arg)
{
	while (!done)
	{
		sub_int_a();
		sub_int_b();
		sub_int_c();
		sub_int_d();
		sub_int_e();
		sub_int_f();
		sub_int_g();
		pthread_mutex_lock (&mutex_reps);
		reps++;
		pthread_mutex_unlock (&mutex_reps);
		
	}
}

// ********************************************************************
// These are the actual integer tests to perform.
// ********************************************************************

// Simple identifiers
void sub_int_a()
{
	int t = 2;
	int weight = 117*ITERATIONS;
	int x1 = 1, x2=-1, x3=-1, x4=-1;
	int i;
	for(i=0; i <= weight; i++)
	{
		x1 = (x1+x2+x3-x4)*t;
		x2 = (x1+x2-x3-x4)*t;
		x3 = (x1-x2+x3+x4)*t;
		x4 = (-x1+x2+x3+x4)*t;
	}
}

// Array elements
void sub_int_b()
{
	int weight = 120*ITERATIONS;
	int t = 2;
	int e1[4] = {1, -1, -1, -1};
	int i;
	for (i=1; i <= weight; i++)
	{
		e1[0] = (e1[0]+e1[1]+e1[2]-e1[3])*t;
		e1[1] = (e1[0]+e1[1]-e1[2]+e1[3])*t;
		e1[2] = (e1[0]-e1[1]+e1[2]+e1[3])*t;
		e1[3] = (-e1[0]+e1[1]+e1[2]+e1[3])*t;
	}
}

// Array as parameter
void sub_int_c()
{
	int weight = 9*ITERATIONS;
	int e1[4] = {1, -1, -1, -1};
	int i;
	for (i=0; i <= weight; i++)
	{
		integer_param_array(e1);
	}
}

// Conditional jumps
void sub_int_d()
{
	int weight = 520*ITERATIONS;
	int cond = 1;
	int i=0;
	for (i=0; i <= weight; i++)
	{
		if (cond == 1)
			cond = 2;
		else
			cond = 3;
		if (cond > 2)
			cond = 0;
		else
			cond = 1;
		if (cond < 1 )
			cond = 1;
		else
			cond = 0;
	}
}

// Integer arithmetic
void sub_int_e()
{
	int weight = 100*ITERATIONS;
	int val1 = 1;
	int val2 = 2;
	int val3 = 3;
	int e1[4];
	int i;
	for (i=0; i <= weight; i++)
	{
		val1 = val1*(val2-val1)*(val3-val2);
		val2 = val3*val2-(val3-val1)*val2;
		val3 = (val3-val2)*(val2+val1);
		e1[val3-2] = val1+val2+val3;
		e1[val2-2] = val1*val2*val3;
	}
}

// Procedure calls
void sub_int_f()
{
	int weight = 95*ITERATIONS;
	int x = 1, y =1, z = 1;
	int i;
	for (i=0; i <= weight; i++)
	{
		integer_proc(x, y, &z);
	}
}

// Standard functions
void sub_int_g()
{
	int weight = 94*ITERATIONS;
	int x = 2;
	int y = 6;
	int t;
	int i;
	for (i=0; i <= weight; i++)
	{
		t = 0;
		x = (t++)*3;
		t = ((t+2)^x)%7;
		x = (t << 2)/--y;
		y += 1;
	}
}

// ********************************************************************
// A function to be used in a test
// ********************************************************************
void integer_proc (int x, int y, int *z)
{
	int t = 5, t2 = 3;
	x  = t*(x+y);
	y  = t*(x+y);
	*z = (x+y)/t2;
}

// ********************************************************************
// A function to be used in a test (Calling with integer arrays)
// ********************************************************************
void integer_param_array(int e[4])
{
	register int j;
	int t = 2, u = 3;

	for (j=0; j<6; j++)
	{
		e[0] = (e[0]+e[1]+e[2]-e[3])*t;
		e[1] = (e[0]+e[1]-e[2]+e[3])*t;
		e[2] = (e[0]-e[1]+e[2]+e[3])*t;
		e[3] = (-e[0]+e[1]+e[2]+e[3])/u;
	}
}
